#include <stdint.h>
#include <stdio.h>
#include "x264.h"
#include "common/cpu.h"
#include "common/pixel.h"
#include "common/i386/pixel.h"

static inline uint64_t read_time(void)
{
    if(sizeof(long)==8)
    {
        uint64_t a, d;
        asm volatile( "rdtsc\n\t" : "=a" (a), "=d" (d) );
        return (d << 32) | (a & 0xffffffff);
    } else {
        uint64_t l;
        asm volatile( "rdtsc\n\t" : "=A" (l) );
        return l;
    }
}

int dezis;
int nop;

#define START_TIMER \
uint64_t tend;\
uint64_t tstart= read_time();

#define STOP_TIMER(N) {\
tend= read_time();\
{\
    static uint64_t tsum=0;\
    static int tcount=0;\
    static int tskip_count=0;\
    if(tcount<2 || tend - tstart < 8*tsum/tcount){\
        tsum+= tend - tstart;\
        tcount++;\
    }else\
        tskip_count++;\
    if(tcount+tskip_count == N){\
        dezis = tsum*10/tcount;\
        tsum = tcount = tskip_count = 0;\
    }\
}}

#define X10(x) x x x x x x x x x x

static uint8_t __attribute__((aligned(16))) buf[0x1100];

static void bench(const char *name, x264_pixel_cmp_t cmp)
{
    int i, j;
    int sum_page=0, sum_line=0, sum_mid=0;
    uint8_t *pix = (uint8_t*)(((intptr_t)buf+0xfff)&~0xfff)-16;
    int speeds[0x90];
    for(i=0; i<0x90; i++)
    {
        for(j=0; j<0x1000; j++)
        {
            START_TIMER;
            X10(cmp(buf, 0, pix+i, 0);)
            STOP_TIMER(0x1000);
        }
        speeds[i] = dezis;
    }
    for(i=0; i<0x10; i++)
        sum_page += speeds[i];
    for(; i<0x80; i++)
        sum_mid += speeds[i];
    for(; i<0x90; i++)
        sum_line += speeds[i];
    printf("%s: avg %d\n", name, (sum_page + 32*sum_mid + 31*sum_line + 0x800) / 0x1000);
    for(i=0; i<0x90; i++)
        printf("%5d%c", (speeds[i]-nop+5)/10, ((i&7)==7)?'\n':' ');
    printf("\n");
}

int main(int argc, char **argv)
{
    int i;
    for(i=0; i<0x1000; i++)
    {
        START_TIMER;
        STOP_TIMER(0x1000);
    }
    nop = dezis;
    printf("nop: %d\n\n", nop);
    int cpu = x264_cpu_detect();

    if(cpu & X264_CPU_MMXEXT) bench("movq8", x264_pixel_sad_8x16_mmxext);
#ifdef __i386__
    if(cpu & X264_CPU_MMXEXT) bench("psllq8", x264_pixel_sad_8x16_palignr_mmxext);
#endif
    if(cpu & X264_CPU_SSSE3)  bench("palignr8", x264_pixel_sad_8x16_palignr_ssse3);

    if(cpu & X264_CPU_MMXEXT) bench("movq16", x264_pixel_sad_16x16_mmxext);
    if(cpu & X264_CPU_SSE2)   bench("movdqu", x264_pixel_sad_16x16_sse2);
    if(cpu & X264_CPU_SSE3)   bench("lddqu", x264_pixel_sad_16x16_sse3);
#ifdef __i386__
    if(cpu & X264_CPU_SSE2)   bench("pslldq", x264_pixel_sad_16x16_palignr_sse2);
#endif
    if(cpu & X264_CPU_SSSE3)  bench("palignr16", x264_pixel_sad_16x16_palignr_ssse3);
    return 0;
}
